#!/usr/bin/env python
# -*- coding: iso-8859-1 -*-

## Copyright 2002-2004 by LivingLogic AG, Bayreuth/Germany.
## Copyright 2002-2004 by Walter Drwald
##
## All Rights Reserved
##
## Permission to use, copy, modify, and distribute this software and its documentation
## for any purpose and without fee is hereby granted, provided that the above copyright
## notice appears in all copies and that both that copyright notice and this permission
## notice appear in supporting documentation, and that the name of LivingLogic AG or
## the author not be used in advertising or publicity pertaining to distribution of the
## software without specific, written prior permission.
##
## LIVINGLOGIC AG AND THE AUTHOR DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
## INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
## LIVINGLOGIC AG OR THE AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
## DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
## IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR
## IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.


"""
<par>This module implements <app>make</app> functionality.</par>

<par>Like <app>make</app> it allows you to specify dependencies
between files and actions to be executed when files don't exist or are out of date
with respect to one of their sources. But unlike <app>make</app> you
can do this in a object oriented way and targets are not only limited to files,
but you can implement e.g. dependencies on database records by implementing new
<pyref class="Target"><class>Target</class></pyref> classes.</par>

<par>Relevant classes are:</par>
<ulist>
<item><pyref class="Project"><class>Project</class></pyref>, which is the container for all targets,
actions and dependencies in a project,</item>
<item><pyref class="Target"><class>Target</class></pyref> (and subclasses), which are files
(or other entities like database records) that have to be built by actions,</item>
<item><pyref class="Dep"><class>Dep</class></pyref> (and subclasses) which specify types
of dependencies between targets and</item>
<item><pyref class="Action"><class>Action</class></pyref> (and subclasses) which specify how targets
are (re)built if they don't exist or aren't up to date or how they should be modified.</item>
</ulist>
"""

__version__ = "$Revision: 1.136 $"[11:-1]

import sys, os, optparse, warnings, re, datetime, cStringIO, errno, tempfile

from ll import ansistyle, url

def dtnow():
	"""
	Return the current time (in &utc;).
	"""
	return datetime.datetime.utcnow()


###
### exceptions & warnings
###

class MakeWarning(Warning):
	"""
	Base class for all warnings in <app>make</app>.
	"""


class RedefinedTargetWarning(MakeWarning):
	"""
	Warning that will be issued when a target is added to a project and a target
	with the same id already exists
	"""

	def __init__(self, targetid):
		self.targetid = targetid

	def __str__(self):
		return "target with id=%r redefined" % self.targetid


class ClockSkewWarning(MakeWarning):
	"""
	Warning that will be issued when a target has a last modification date from the future.
	"""

	def __init__(self, targetid, lastmodified, now):
		self.targetid = targetid
		self.lastmodified = lastmodified
		self.now = now

	def __str__(self):
		return "target %r is from the future: %r>%r=now" % (self.targetid, self.lastmodified, self.now)


class UndefinedTargetError(KeyError):
	"""
	Exception that will be raised, when a target with the specified id doesn't exist within the project.
	"""

	def __init__(self, id):
		self.id = id

	def __str__(self):
		return "target %r undefined" % self.id


class InputError(Exception):
	"""
	Exception that will be raised, when a target has the wrong number of inputs.
	"""

	def __init__(self, action, target, inputs):
		self.action = action
		self.target = target
		self.inputs = inputs

	def __str__(self):
		return "wrong number of inputs for target %r in action %r: %r" % (self.target, self.action, self.inputs)


class ChainError(Exception):
	"""
	Exeption that will be raised, when an action in a chain gets an illegal input
	"""


###
### Dependency classes
###

class Dep(object):
	"""
	<par>A <class>Dep</class> is a dependency between an <z>input</z>
	and an <z>output</z> <pyref class="Target"><class>Target</class></pyref>.
	It's possible to assign specific meanings to certain dependencies by using
	subclasses of <class>Dep</class>.</par>
	"""
	__slots__ = ("input", "output")

	def __init__(self, input, output):
		"""
		<par>create a <class>Dep</class> instance.</par>
		"""
		self.input = input
		self.output = output
		output._inputs.append(self)
		input._outputs.append(self)

	def __hash__(self):
		return hash(self.input.id) ^ hash(self.output.id)

	def __eq__(self, other):
		return self.input.id == other.input.id and self.output.id == other.output.id

	def __ne__(self, other):
		return self.input.id != other.input.id or self.output.id != other.output.id

	def __repr__(self):
		return "<%s from %r to %r at 0x%x>" % (self.__class__.__name__, self.input, self.output, id(self))


class MainDep(Dep):
	"""
	<par>This class can be used to create a special dependency between
	two targets. This is used by <pyref class="ReadAction"><class>ReadAction</class></pyref>
	to read the main input target.</par>
	"""
	pass


class OrderedDep(MainDep):
	"""
	<par>This dependency class represents an ordered dependency, i.e. each
	instance has an <lit>order</lit> attribute.</par>
	"""
	order = 0

	__slots__ = ("order")

	def __init__(self, source, target, order=None):
		"""
		<par>create a new <class>OrderedDep</class> instance.
		When <arg>order</arg> is not specified a new unique order value
		will be used that will be greater than the order value of
		previously created <class>OrderedDep</class>
		instances.</par>
		"""
		MainDep.__init__(self, source, target)
		if order is None:
			order = self.__class__.order
			self.__class__.order += 1
		self.order = order


###
### Target classes
###

class Target(object):
	def __init__(self, project, id, *actions):
		"""
		<par>Create a new <class>Target</class> instance.</par>
		<par><arg>actions</arg> can must be list of <pyref class="Action"><class>Action</class></pyref> object,
		which will be executed sequencially to create the target</par>
		"""
		self.project = project
		self.id = id
		self.actions = actions
		self._inputs = []
		self._outputs = []
		self.project[self.id] = self

	def lastmodified(self):
		"""
		<par>return the timestamp, when this target was last modified or <lit>None</lit>
		when the target does not exist. The timestamp will be cached and will be updated
		when a target is remade.</par>
		"""
		try:
			return self._lastmodified
		except AttributeError:
			lm = self._fetchlastmodified()
			now = dtnow()
			if lm is not None and lm>now:
				self.project.reportWarning(ClockSkewWarning(self, lm, now), 2)
			self._lastmodified = lm
			return lm

	def _fetchlastmodified(self):
		"""
		<par>return the timestamp of the last modification.
		Overwrite in subclasses.</par>
		"""
		raise NotImplementedError()

	def touch(self):
		"""
		<par>set the <z>last modified</z> timestamp to now.</par>
		"""
		self._lastmodified = dtnow()

	def dirty(self):
		"""
		<par>forget the cached <z>last modifified</z> timestamp,
		so the target might be rebuilt on subsequent calls.</par>
		"""
		try:
			del self._lastmodified
		except AttributeError:
			pass

	def clear(self):
		"""
		<par>Called by <pyref class="Project" method="clear"><method>Project.clear</method></pyref>
		to get rid of all the old targets.</par>
		"""
		self._inputs = []
		self._outputs = []

	def _reportinputsoutputs(self):
		if self.project.showoutputs:
			if self.project.showlevel:
				self.project.report(self.project.TextProgress(self.project.level*self.project.indent))
			self.project.report(self.project.strtarget(self))
			if self.project.showinputs:
				inputs = self.inputs(MainDep)
				if len(inputs):
					if len(inputs)>self.project.maxinputreport:
						reason = self.TextTarget("(%d inputs)" % len(inputs))
					else:
						reason = ansistyle.Text()
						for input in inputs:
							if reason:
								reason.append("+")
							reason.append(self.project.strtarget(input))
				else:
					reason = None
				if reason is not None:
					self.project.report(self.project.TextProgress("<--", reason))
			self.project.report(":")

	def _reportaction(self, action):
		if self.project.showactions:
			self.project.report(" ", self.project.straction(action))

	def _reporttimedata(self, data, time):
		if self.project.showactions:
			if self.project.showtime:
				self.project.report(self.project.TextProgress("=", self.project.strseconds(time)))
			if self.project.showdata and data is not None:
				self.project.report(self.project.TextProgress(" ["))
				if isinstance(data, str):
					name = self.project.strsize(len(data))
				else:
					name = type(data)
					if name.__module__ != "__builtin__":
						name = "%s.%s" % (name.__module__, name.__name__)
					else:
						name = data.__name__
					name = self.project.TextData(name)
				self.project.report(name)
				self.project.report(self.project.TextProgress("]"))

	def _reportdone(self, inputtime, selftime, showinput):
		if self.project.showoutputs:
			self.project.report(" ", self.project.TextDone("done"))
			if self.project.showtime:
				if showinput:
					self.project.report(self.project.TextProgress("=", self.project.strseconds(inputtime), "+", self.project.strseconds(selftime)))
				else:
					self.project.report(self.project.TextProgress("=", self.project.strseconds(selftime)))

	def update(self):
		"""
		<par>Update this target and all of its sources, if necessary.</par>
		"""
		self.project.level += 1
		t0 = dtnow()
		inputactions = 0
		for dep in self._inputs:
			inputactions += len(dep.input.actions)
			dep.input.update()
		if self.hasToBeRebuilt():
			try:
				self._reportinputsoutputs()
				failed = False
				try:
					data = None
					t1 = dtnow()
					for action in self.actions:
						t2 = dtnow()
						self._reportaction(action)
						data = action.execute(self, data)
						t3 = dtnow()
						self._reporttimedata(data, t3-t2)
					t4 = dtnow()
					self._reportdone(t1-t0, t4-t1, inputactions)
				except SystemExit:
					raise
				except KeyboardInterrupt:
					raise
				except Exception, ex:
					if self.project.ignoreerrors:
						failed = True
					else:
						raise
				t2 = dtnow()
				if failed:
					import traceback
					self.project.reportError(traceback.format_stack())
					self.project.reportError("\n", self.project.strerror("failed"), " after ", self.project.strseconds(t2-t1))
					self.project.targetsfailed += 1
				else:
					self.project.targetsrebuilt += 1
			finally:
				self.project.report("\n")
			self.touch() # always touch the target, so building can continue even if this target failed
		self.project.level -= 1

	def iterinputs(self, type=Dep):
		"""
		<par>Iterate through all input targets of <self/> where the dependency is a subtype
		of <arg>type</arg>.</par>
		"""
		for dep in self._inputs:
			if isinstance(dep, type):
				yield dep.input

	def inputs(self, type=Dep):
		"""
		<par>return a list of all the input targets for <self/> where the dependency is a subtype
		of <arg>type</arg>.</par>
		"""
		return list(self.iterinputs(type))

	def iterinputdeps(self, type=Dep):
		"""
		<par>Iterate through all the input dependancies for <self/> that are a subtype
		of <arg>type</arg>.</par>
		"""
		for dep in self._inputs:
			if isinstance(dep, type):
				yield dep

	def inputdeps(self, type=Dep):
		"""
		<par>return a list of all the input dependancies for <self/> that are a subtype
		of <arg>type</arg>.</par>
		"""
		return list(self.iterinputdeps(type))

	def iteroutputs(self, type=Dep):
		"""
		<par>Iterate through all output targets of <self/> where the dependency is a subtype
		of <arg>type</arg>.</par>
		"""
		for dep in self._outputs:
			if isinstance(dep, type):
				yield dep.output

	def outputs(self, type=Dep):
		"""
		<par>return a list of all the target dependancies for <self/> that are a subtype
		of <arg>type</arg>.</par>
		"""
		return list(self.iteroutputs(type))

	def iteroutputdeps(self, type=Dep):
		"""
		<par>Iterate through all the output dependancies for <self/> that are a subtype
		of <arg>type</arg>.</par>
		"""
		for dep in self._outputs:
			if isinstance(dep, type):
				yield dep

	def outputdeps(self, type=Dep):
		"""
		<par>return a list of all the output dependencies for <self/> where the dependency is a subtype
		of <arg>type</arg>.</par>
		"""
		return list(self.iteroutputdeps(type))

	def newerinputdeps(self, type=Dep):
		"""
		<par>return a list of all the input targets that or newer the this one
		(or don't exist).</par>
		"""
		selfLM = self.lastmodified()
		inputs = self.inputdeps(type)
		if selfLM is None:
			return inputs
		else:
			newer = []
			for dep in inputs:
				inputLM = dep.input.lastmodified()
				if inputLM is None or inputLM>selfLM:
					newer.append(dep)
			return newer

	def newersources(self, type=Dep):
		"""
		<par>return a list of all the input targets that or newer the this one
		(or don't exist).</par>
		"""
		return [ dep.input for dep in self.newerinputdeps(type) ]

	def dependOn(self, *others):
		"""
		<par>add one or more dependencies.</par>

		<par><arg>others</arg> may contain <pyref class="Target"><class>Target</class></pyref>
		instances and subclasses of <pyref class="Dep"><class>Dep</class></pyref>. This will be used to
		create specific dependencies. Initially <pyref class="Dep"><class>Dep</class></pyref> will be used to
		construct input dependencies for every <pyref class="Target"><class>Target</class></pyref> instance
		encountered in the argument list. From the point in the argument list where a
		<pyref class="Dep"><class>Dep</class></pyref> subclass is encountered this new class will be
		used for constructing dependencies. This makes it possible to switch between different
		dependency classes.</par>
		"""
		dependencyClass = Dep
		for other in others:
			if isinstance(other, type) and issubclass(other, Dep):
				dependencyClass = other
			else:
				dependency = dependencyClass(other, self)

	def hasToBeRebuilt(self):
		"""
		<par>does <self/> have to be rebuilt (either because this target does not exist, or
		it is older than one of its sources)?</par>
		"""
		lm = self.lastmodified()
		if lm is not None:
			return len(self.newerinputdeps())>0
		else:
			return True

	def __repr__(self):
		return "<%s object id=%r at 0x%x>" % (self.__class__.__name__, self.id, id(self))

	def failed(self):
		"""
		Called by actions when building a target fails when writing the output
		has already begun. Derived classes will overwrite <method>failed</method>
		to remove broken files.
		"""
		self.project.report("Removing broken target %s ...\n" % self)


class FileSystemTarget(Target):
	"""
	<par>A target in the filesystem, i.e. a file or directory.</par>
	"""
	pass


class FileTarget(FileSystemTarget):
	"""
	<par>a file</par>
	"""
	name = "File"

	def _fetchlastmodified(self):
		try:
			return self.id.mtime()
		except OSError:
			return None

	def failed(self):
		super(FileTarget, self).failed()
		if self.id.exists():
			self.id.remove()


class DirectoryTarget(FileSystemTarget):
	"""
	<par>a directory.</par>
	"""
	name = "Dir"

	def _fetchlastmodified(self):
		# only rebuild if it doesn't exist, if it exists do nothing
		try:
			timestamp = self.id.mtime()
		except OSError:
			return None
		else:
			return datetime.datetime(1900, 1, 1) # should be older than any existing file/directory

	def hasToBeRebuilt(self):
		return self.lastmodified() is None


class ImageTarget(FileTarget):
	"""
	<par>an image file.</par>
	"""
	name = "Image"


class JavascriptTarget(FileTarget):
	"""
	<par>a Javascript source file (extension <lit>.js</lit>)</par>
	"""
	name = "JS"


class CascadingStyleSheetTarget(FileTarget):
	"""
	<par>a Cascading Stylesheet file (extension <lit>.css</lit>)</par>
	"""
	name = "CSS"


class PHPTarget(FileTarget):
	"""
	<par>a PHP include file (extension <lit>.php</lit>)</par>
	"""
	name = "PHP"


class HTMLTarget(FileTarget):
	"""
	<par>a &html; file (extension <lit>.html</lit>, <lit>.phtml</lit>, <lit>.jsp</lit>, etc.)</par>
	"""
	name = "HTML"


class XMLTarget(FileTarget):
	name = "XML"
	"""
	<par>an &xml; file (extension <lit>.xml</lit>)</par>
	"""


class DTDTarget(FileTarget):
	"""
	<par>a Document Type Definition file (extension <lit>.dtd</lit>)</par>
	"""
	name = "DTD"


class PDFTarget(FileTarget):
	"""
	<par>a &pdf; file (extension <lit>.pdf</lit>)</par>
	"""
	name = "PDF"


class JavaPropTarget(FileTarget):
	"""
	<par>a Java properties file (extension <lit>.properties</lit>)</par>
	"""
	name = "JavaProp"


class SQLTarget(FileTarget):
	"""
	<par>an &sql; script (extension <lit>.sql</lit>)</par>
	"""
	name = "SQL"


class PythonTarget(FileTarget):
	"""
	<par>a Python source file (extension <lit>.py</lit>)</par>
	"""
	name = "Python"

	def __init__(self, project, id, *actions):
		FileTarget.__init__(self, project, id, *actions)
		self.module = None

	def clear(self):
		# clearing the import cache is the job of Project.build()
		self.module = None
		FileTarget.clear(self)

	def dirty(self):
		# clearing the import cache is the job of Project.build()
		self.module = None
		FileTarget.dirty(self)

	def load(self, mode="new"):
		"""
		<par>import and return the Python module as the result.</par>
		"""
		if self.module is None:
			self.module = self.id.import_(mode)
		return self.module


class XISTNSTarget(PythonTarget):
	"""
	<par>a Python source file that contains an &xist; <pyref module="ll.xist.xsc" class="Namespace"><class>Namespace</class></pyref>
	object and must be imported to provide element definitions etc. to the &xist; parser.</par>
	"""
	name = "XISTNS"

	def __init__(self, project, id, prefix=None, *actions):
		PythonTarget.__init__(self, project, id, *actions)
		self.prefix = prefix


class XISTTarget(FileTarget):
	"""
	<par>an &xist; source file, i.e. a &xml; file that has to be transformed via <pyref module="ll.xist">&xist;</pyref>
	(extensions <lit>.xmlxsc</lit>, <lit>.htmlxsc</lit>, etc.)</par>
	"""
	name = "XIST"


class PhonyTarget(Target):
	"""
	<par>A <class>PhonyTarget</class> does not really exist. It can
	be used as a virtual target for all the real targets that have to be built.</par>
	"""
	name = "Phony"

	def __init__(self, project, id, doc=None, *actions):
		"""
		<par>create a new <class>PhonyTarget</class>.</par>

		<par><arg>doc</arg> should be a description of the <class>PhonyTarget</class>
		and will be printed when <pyref class="Project" method="buildwithargs"><method>buildwithargs</method></pyref>
		is called without arguments.</par>
		"""
		Target.__init__(self, project, id, *actions)
		self.doc = doc

	def _fetchlastmodified(self):
		"""
		<par>get the <z>last modification</z> timestamp. This will always
		be <lit>None</lit>, as a <class>PhonyTarget</class> does not really exist.
		Note that this means that the associated actions will be executed
		every time the target is updated.</par>
		"""
		return None


class DBTarget(Target):
	"""
	<par>a database procedure or function</par>
	"""
	name = "DB"

	def _fetchlastmodified(self):
		return None


###
### Action classes
###

class Action(object):
	"""
	<par>An <class>Action</class> is responsible for rebuilding or modifying a
	<pyref class="Target"><class>Target</class></pyref>.</par>

	<par>An <class>Action</class> is stateless and so may be used by multiple targets.
	Executing the action is done through a call to the method
	<pyref method="execute"><method>execute</method></pyref>.</par>

	<par>Creating the final target is done by executing a chain of actions.
	The first action loads the data (usually just the file content) from (one of)
	the source targets and passes the data on to the next action in the chain. The
	next several actions may transform the data in a number of ways until the chain
	reaches a save action. This save action stores the resulting data into the final
	target. Remaining actions in the chain can be used to modify the target (e.g. by
	changing the owner of the file or its permission bits).</par>
	"""
	def __init__(self):
		"""
		<par>create a new <class>Action</class> instance.</par>
		"""
		pass

	def execute(self, target, data=None):
		"""
		<par>execute the action to rebuild or modify <arg>target</arg>. Depending
		on the action itself, the input data is either read from the input targets
		of <arg>target</arg> or <arg>data</arg> is used (which must be produced by
		the previous action in the action chain). Furthermore the action may
		save the result to target, modify the target (e.g. change the owner of the
		file) or return modified data (which will be passed to the next action in
		the chain).</par>
		"""
		raise NotImplementedError()


class ReadAction(Action):
	"""
	<par>Read the content from the only input <pyref class="Target"><class>Target</class></pyref>
	on which the target depends through a <pyref class="MainDep"><class>MainDep</class></pyref>.</par>
	"""
	name = "Read"

	def execute(self, target, data=None):
		"""
		<par>execute the action.</par>
		<par>If <arg>target</arg> doesn't have exactly one main source, an exception
		will be raised.</par>
		"""
		inputs = target.inputs(MainDep)
		if len(inputs) != 1:
			raise InputError(self, target, inputs)

		infile = inputs[0].id.openread()
		data = infile.read()
		infile.close()
		return data


class WriteAction(Action):
	"""
	<par>Write the incoming data to the target.</par>
	"""
	name = "Write"

	def execute(self, target, data=None):
		"""
		<par>Write the string <arg>data</arg> to <arg>target</arg>.</par>
		"""
		if not isinstance(data, basestring):
			raise ChainError("need a string as input")
		outfile = target.id.openwrite()
		outfile.write(data)
		outfile.close()


class JoinedReadAction(Action):
	"""
	<par>Read multiple input targets and join them. The input targets must depend
	on the target via a <pyref class="OrderedDep"><class>OrderedDep</class></pyref>
	and will be joined in the order specified by these dependencies.</par>
	"""
	name = "JoinedRead"

	def execute(self, target, data=None):
		"""
		<par>joins the sources in the order specified by the dependencies.</par>
		"""
		deps = target.inputdeps(OrderedDep)
		deps.sort(lambda s1, s2: cmp(s1.order, s2.order))

		data = []
		for dep in deps:
			infile = dep.input.id.openread()
			data.append(infile.read())
			infile.close()
		return "".join(data)


class MkDirAction(Action):
	"""
	<par>This action creates the target as a directory.</par>
	"""
	name = "MkDir"

	def __init__(self, mode=0777):
		"""
		Create a <class>MkDirAction</class> instance. <arg>mode</arg> (which defaults
		to <lit>0777</lit>) will be used as the permission bit pattern for the new directory.
		"""
		super(MkdDirAction, self).__init__()
		self.mode = mode

	def execute(self, target, data=None):
		"""
		<par>create the directory with the permission bits specified in the constructor.</par>
		"""
		target.id.makedirs(self.mode)

	def __repr__(self):
		return "<%s.%s mode=%s at 0x%x>" % (self.__class__.__module__, self.__class__.__name__, oct(self.mode), id(self))


class XISTParseAction(Action):
	"""
	<par>This action parses a string into an <pyref module="ll.xist"><app>&xist;</app></pyref> node.</par>
	"""
	name = "XISTParse"

	def __init__(self, parser=None, base=None):
		super(XISTParseAction, self).__init__()
		if parser is None:
			from ll.xist import parsers
			parser = parsers.Parser()
		self.parser = parser
		self.base = base

	def loadnamespaces(self, target):
		"""
		<par>import all the source namespaces (i.e. all the input targets of <arg>target</arg>
		that are instances of <pyref class="XISTNSTarget"><class>XISTNSTarget</class></pyref>).</par>
		<par>The prefixes and module objects will be returned in a list of tuples.</par>
		"""
		return [ (input.prefix, input.load()) for input in target.iterinputs() if isinstance(input, XISTNSTarget) ]

	def execute(self, target, data=None):
		"""
		Parse the string <arg>data</arg> into an &xml; node. In addition to the
		namespaces available for parsing in <lit><self/>.parser</lit> all namespaces
		loaded by <pyref method="loadnamespaces"><method>loadnamespaces</method></pyref>
		will be used.
		"""
		namespaces = self.loadnamespaces(target)
		if namespaces: # We need additional namespaces
			oldprefixes = self.parser.prefixes # Remember old prefixes
			prefixes = oldprefixes.clone()
			for (prefix, ns) in namespaces:
				prefixes[prefix].insert(0, ns)
			try:
				self.parser.prefixes = prefixes
				node = self.parser.parseString(data, self.base)
			finally:
				self.parser.prefixes = oldprefixes # Restore old prefixes
		else:
			node = self.parser.parseString(data, self.base)
		return node

	def __repr__(self):
		options = []
		for optionname in ("parser", "base"):
			optionvalue = getattr(self, optionname)
			if optionvalue is not None:
				options.append(" %s=%r" % (optionname, optionvalue))
		return "<%s.%s%s at 0x%x>" % (self.__class__.__module__, self.__class__.__name__, "".join(options), id(self))


class XISTConvertAction(Action):
	"""
	<par>This action transforms an <pyref module="ll.xist"><app>&xist;</app></pyref> node.</par>
	"""
	name = "XISTConvert"

	def __init__(self, mode=None, target=None, stage=None, lang=None, targetroot=None):
		"""
		<par>create a new <class>XISTAction</class> instance. The arguments
		will be used the create a <pyref module="ll.xist.converters" class="Converter"><class>Converter</class></pyref>
		for each call to <method>execute</method>.</par>
		"""
		super(XISTConvertAction, self).__init__()
		self.mode = mode
		self.target = target
		self.stage = stage
		self.lang = lang
		self.targetroot = targetroot

	def converter(self, maketarget=None):
		"""
		<par>create a new <pyref module="ll.xist.converters" class="Converter"><class>Converter</class></pyref>
		instance to be used by this action. The attributes of this new converter (<lit>mode</lit>, <lit>target</lit>,
		<lit>stage</lit>, etc.) will correspond to those specified in the constructor.</par>
		<par>When the argument <arg>maketarget</arg> is specified it will be used to initialize the
		attribute <lit>maketarget</lit> of the converter.</par>
		"""
		from ll.xist import converters
		return converters.Converter(root=self.targetroot, mode=self.mode, stage=self.stage, target=self.target, lang=self.lang, makeaction=self, maketarget=maketarget)

	def execute(self, target, data=None):
		"""
		<par>convert the &xist; node <arg>data</arg> using a converter provided
		by <pyref method="converter"><method>converter</method></pyref>.</par>
		"""
		from ll.xist import xsc
		if not isinstance(data, xsc.Node):
			raise ChainError("need an XIST node as input")
		return data.convert(self.converter(target))

	def __repr__(self):
		options = []
		for optionname in ("mode", "target", "stage", "lang", "targetroot"):
			optionvalue = getattr(self, optionname)
			if optionvalue is not None:
				options.append(" %s=%r" % (optionname, optionvalue))
		return "<%s.%s%s at 0x%x>" % (self.__class__.__module__, self.__class__.__name__, "".join(options), id(self))


class XISTPublishAction(Action):
	"""
	<par>This action publishes an <pyref module="ll.xist"><app>&xist;</app></pyref> node as a string.</par>
	"""
	name = "XISTPublish"

	def __init__(self, publisher=None, base=None):
		super(XISTPublishAction, self).__init__()
		if publisher is None:
			from ll.xist import publishers
			publisher = publishers.Publisher()
		self.publisher = publisher
		self.base = base

	def execute(self, target, data=None):
		from ll.xist import xsc
		if not isinstance(data, xsc.Node):
			raise ChainError("need an XIST node as input")
		stream = cStringIO.StringIO()
		self.publisher.dopublication(stream, data, self.base)
		return stream.getvalue()

	def __repr__(self):
		options = []
		for optionname in ("publisher", "base"):
			optionvalue = getattr(self, optionname)
			if optionvalue is not None:
				options.append(" %s=%r" % (optionname, optionvalue))
		return "<%s.%s%s at 0x%x>" % (self.__class__.__module__, self.__class__.__name__, "".join(options), id(self))


class XISTTextAction(Action):
	"""
	<par>This action returns a text version of an <pyref module="ll.xist"><app>&xist;</app></pyref> node.</par>
	"""
	name = "XISTText"

	def execute(self, target, data=None):
		from ll.xist import xsc
		if not isinstance(data, xsc.Node):
			raise ChainError("need an XIST node as input")
		return data.asText()


class FOPAction(Action):
	"""
	Transforms an &xml; string (containing XSL-FO) into &pdf; (requires Apache FOP)
	"""
	name = "FOP"
	command = "/usr/local/src/fop-0.20.5/fop.sh -q -c /usr/local/src/fop-0.20.5/conf/userconfig.xml -fo %s -pdf %s"

	def execute(self, target, data=None):
		(infd, inname) = tempfile.mkstemp(suffix=".fo")
		(outfd, outname) = tempfile.mkstemp(suffix=".pdf")
		try:
			infile = os.fdopen(infd, "wb")
			os.fdopen(outfd).close()
			infile.write(data)
			infile.close()
			os.system(self.command % (inname, outname))
			data = open(outname, "rb").read()
		finally:
			os.remove(inname)
			os.remove(outname)
		return data


class TOXICAction(Action):
	"""
	<par>This action transforms an &xml; string into an Oracle procedure body via
	<pyref module="ll.toxic"><module>ll.toxic</module></pyref>.</par>
	"""
	name = "TOXIC"

	def __init__(self, encoding="us-ascii"):
		super(TOXICAction, self).__init__()
		self.encoding = encoding

	def execute(self, target, data=None):
		if not isinstance(data, basestring):
			raise ChainError("need a string as input")
		from ll import toxic
		return toxic.xml2ora(data, self.encoding)


class TOXICPrettifyAction(Action):
	"""
	<par>This action tries to fix the indentation of a PL/SQL snippet via
	<pyref module="ll.toxic" function="prettify"><function>prettify</function></pyref>.</par>
	"""
	name = "TOXICPrettify"

	def execute(self, target, data=None):
		if not isinstance(data, basestring):
			raise ChainError("need a string as input")
		from ll import toxic
		return toxic.prettify(data)


class SplatAction(Action):
	"""
	<par>This action transforms an input string by replacing certain regular expressions.</par>
	"""
	name = "Splat"

	def __init__(self, patterns):
		"""
		<par>create a new <class>SplatAction</class> instance. <arg>patterns</arg>
		is a list of pattern pairs. The first entry will be replaced by the second entry.</par>
		"""
		super(SplatAction, self).__init__()
		self.patterns = patterns

	def execute(self, target, data=None):
		if data is None:
			raise ChainError("need an string as input")
		for (search, replace) in self.patterns:
			data = re.sub(search, replace, data)
		return data

	def __repr__(self):
		return "<%s.%s patterns=%r at 0x%x>" % (self.__class__.__module__, self.__class__.__name__, self.patterns, id(self))

	def __str__(self):
		return "Splat(%r)" % (self.patterns,)


class CommandAction(Action):
	"""
	<par>This action executes a system command (via <function>os.system</function>).</par>
	"""
	name = "Command"

	def __init__(self, command):
		"""
		<par>create a new <class>CommandAction</class>. <arg>command</arg> is the command
		that will executed when <pyref method="execute"><method>execute</method></pyref> is called.</par>
		"""
		super(CommandAction, self).__init__()
		self.command = command

	def execute(self, target, data=None):
		"""
		<par>execute the action (note that the <arg>target</arg> is irrelavant and will be ignored;
		in most cases it will probably be a <pyref class="PhonyTarget"><class>PhonyTarget</class></pyref>).</par>
		"""
		os.system(self.command)

	def __repr__(self):
		return "<%s.%s command=%r at 0x%x>" % (self.__class__.__module__, self.__class__.__name__, self.command, id(self))

	def __str__(self):
		return "%s(%r)" % (super(CommandAction, self).__str__(), self.command)


class ModeAction(Action):
	"""
	<class>ModeAction</class> is used for changing file permissions.
	"""
	name = "Mode"

	def __init__(self, mode=0644):
		"""
		Create an <class>ModeAction</class> instance. <arg>mode</arg> (which defaults to <lit>0644</lit>)
		will be use as the permission bit pattern.
		"""
		super(ModeAction, self).__init__()
		self.mode = mode

	def execute(self, target, data=None):
		target.id.chmod(self.mode)

	def __repr__(self):
		return "<%s.%s mode=%s at 0x%x>" % (self.__class__.__module__, self.__class__.__name__, oct(self.mode), id(self))

	def __str__(self):
		return "%s(%s)" % (super(ModeAction, self).__str__(), oct(self.mode))


class OwnerAction(Action):
	"""
	<class>OwnerAction</class> is used for changing the user and/or group ownership of files.
	"""
	name = "Owner"

	def __init__(self, user=None, group=None):
		"""
		Create a new <class>OwnerAction</class> instance. <arg>user</arg> can either be a numerical
		user id or a user name or <lit>None</lit>. If it is <lit>None</lit> no user ownership will
		be changed. The same applies to <arg>group</arg>.
		"""
		super(OwnerAction, self).__init__()
		self.user = user
		self.group = group

	def execute(self, target, data=None):
		target.id.chown(self.user, self.group)

	def __repr__(self):
		v = []
		if self.user is not None:
			v.append(" user=%r" % self.user)
		if self.group is not None:
			v.append(" group=%r" % self.group)
		return "<%s.%s%s at 0x%x>" % (self.__class__.__module__, self.__class__.__name__, "".join(v), id(self))

	def __str__(self):
		v = []
		if self.user is not None:
			v.append("user=%r" % self.user)
		if self.group is not None:
			v.append("group=%r" % self.group)
		return "%s(%s)" % (super(OwnerAction, self).__str__(), ", ".join(v))


###
### Classes for target ids (apart from strings for PhonyTargets and URLs for file targets)
###

class DBID(object):
	"""
	<par>This class provides a unique identifier for database content. This
	can be used as an id for <pyref class="Target"><class>Target</class></pyref>
	objects that are not files, but database records, function, procedures etc.</par>
	"""
	name = None

	def __init__(self, connection, type, name, key=None):
		"""
		<par>Create a new <class>DBID</class> instance. Arguments are:</par>
		<dlist>
		<term><arg>connection</arg></term>
		<item>A string that specifies the connection to the database.
		E.g. <lit>"user/pwd@db.host.com"</lit> for Oracle.</item>
		<term><arg>type</arg></term>
		<item>The type of the object. Values may be <lit>"table"</lit>,
		<lit>"view"</lit>, <lit>"function"</lit>, <lit>"procedure"</lit> etc.</item>
		<term><arg>name</arg></term>
		<item>The name of the object</item>
		<term><arg>key</arg></term>
		<item>If <arg>name</arg> refers to a table, <arg>key</arg> can be used
		to specify a row in this table.</item>
		</dlist>
		"""
		self.connection = connection
		self.type = type
		self.name = name
		self.key = key

	def __eq__(self, other):
		res = self.__class__ == other.__class__
		if not res:
			res = self.connection==other.connection and self.type==other.type and self.name==other.name and self.key==other.key
		return res

	def __hash__(self):
		return hash(self.connection) ^ hash(self.type) ^ hash(self.name) ^ hash(self.key)

	def __repr__(self):
		args = []
		for attrname in ("connection", "type", "name", "key"):
			attrvalue = getattr(self, attrname)
			if attrvalue is not None:
				args.append("%s=%r" % (attrname, attrvalue))
		return "%s(%s)" % (self.__class__.__name__, ", ".join(args))

	def __str__(self):
		s = "%s:%s|%s:%s" % (self.__class__.name, self.connection, self.type, self.name)
		if self.key is not None:
			s += "|%s" % (self.key,)
		return s


class OracleID(DBID):
	name = "oracle"

	def openwrite(self):
		return OracleWriteResource(self)

	def openread(self):
		return OracleReadResource(self)


class OracleResource(object):
	def __init__(self, loc):
		self.loc = loc

	def _cursor(self):
		"""
		Create a database cursor for this connection
		"""
		import cx_Oracle
		db = cx_Oracle.connect(self.loc.connection)
		return db.cursor()

	def __repr__(self):
		return "<%s instance for %r at 0x%x>" % (self.__class__.__name__, self.loc, id(self))


class OracleWriteResource(OracleResource):
	def __init__(self, loc):
		OracleResource.__init__(self, loc)
		self.stream = None

	def __getattr__(self, attrname):
		if self.stream is None:
			self.stream = cStringIO.StringIO()
			if self.loc.type in ("function", "procedure"):
				self.stream.write("create or replace %s %s\n" % (self.loc.type, self.loc.name))
			else:
				raise ValueError("don't know how to handle %r" % self.loc)
		return getattr(self.stream, attrname)

	def __del__(self):
		self.close()

	def close(self):
		if self.stream is not None:
			c = self._cursor()
			c.execute(self.stream.getvalue())
			self.stream = None


class OracleReadResource(OracleResource):
	def __init__(self, loc):
		OracleResource.__init__(self, loc)
		self.stream = None

	def __getattr__(self, attrname):
		if self.stream is None:
			if self.loc.type not in ("function", "procedure"):
				raise ValueError("don't know how to handle %r" % self.loc)
			c = self._cursor()
			c.execute("select text from user_source where lower(name)=lower(:name) and type='%s' and line>1 order by line" % self.loc.type.upper(), name=self.loc.name)
			rows = c.fetchall()
			if not rows:
				raise IOError(errno.ENOENT, "no such %s: %s" % (self.loc.type, self.loc.name))
			self.stream = cStringIO.StringIO("".join([row[0] for row in rows]))
		return getattr(self.stream, attrname)


###
### The project class
###

class Project(dict):
	"""
	<par>A <class>Project</class> collects all <pyref class="Target"><class>Target</class></pyref>s
	from a project. It is responsible for initiating the build process and for generating a report about
	the progress of the build process.</par>
	"""
	def __init__(self, showsummary=True, showoutputs=True, showinputs=False, showlevel=True, showactions=True, showdata=False, showtime=False):
		super(Project, self).__init__()
		self.targetsrebuilt = 0
		self.targetsfailed = 0
		self.showsummary = showsummary
		self.showoutputs = showoutputs
		self.showinputs = showinputs
		self.showlevel = showlevel
		self.showactions = showactions
		self.showdata = showdata
		self.showtime = showtime
		self.ignoreerrors = False
		color = os.getenv("LL_MAKE_REPRANSI")
		if color is None:
			color = 0
		else:
			color = int(color)
			if not sys.platform.startswith("linux") and not sys.platform.startswith("darwin"):
				color = 0
		self.color = color
		self._initcolors()
		self.maxinputreport = 2
		self.__built = False # remember whether build() has been called, so we can reset timestamps on the second call
		self.here = None # cache the current directory during builds (used for shortening URLs)
		self.home = None # cache the home directory during builds (used for shortening URLs)
		self.level = None # keep track of the recursion depth during calls to build()
		self.indent = os.getenv("LL_MAKE_INDENT", "> ") # indentation for recursion tracking

	def __repr__(self):
		return "<%s.%s with %d targets at 0x%x>" % (self.__module__, self.__class__.__name__, len(self), id(self))

	def _initcolors(self):
		"""
		<par>initialize the color classes which are used for the various formatting
		methods.</par>
		"""
		if self.color:
			self.TextProgress = ansistyle.Text8
			self.TextTarget = ansistyle.Text7
			self.TextID = ansistyle.Text2
			self.TextAction = ansistyle.Text4
			self.TextTime = ansistyle.Text5
			self.TextData = ansistyle.Text6
			self.TextSize = ansistyle.Text6
			self.TextDone = ansistyle.Text15
			self.TextError = ansistyle.Text9
		else:
			self.TextProgress = ansistyle.Text
			self.TextTarget = ansistyle.Text
			self.TextID = ansistyle.Text
			self.TextAction = ansistyle.Text
			self.TextTime = ansistyle.Text
			self.TextData = ansistyle.Text
			self.TextSize = ansistyle.Text
			self.TextDone = ansistyle.Text
			self.TextError = ansistyle.Text

	def strseconds(self, delta):
		"""
		<par>return a nicely formatted and colored string for
		the <class>datetime.timedelta</class> value <arg>delta</arg>.</par>
		"""
		rest = delta.seconds

		(rest, secs) = divmod(rest, 60)
		(rest, mins) = divmod(rest, 60)
		rest += delta.days*24

		secs += delta.microseconds/1000000.
		if rest:
			text = "%d:%02d:%06.3fh" % (rest, mins, secs)
		elif mins:
			text = "%02d:%06.3fm" % (mins, secs)
		else:
			text = "%.3fs" % secs
	
		return self.TextTime(text)

	def strsize(self, bytes):
		"""
		<par>return a nicely formatted and colored string for
		the byte count value <arg>bytes</arg>.</par>
		"""
		return self.TextSize("%db" % bytes)

	def straction(self, action):
		"""
		<par>return a nicely formatted and colored string for
		the action <arg>action</arg>.</par>
		"""
		try:
			name = action.__class__.name
		except AttributeError:
			name = action.__class__.__name__

		return self.TextAction(name)

	def strerror(self, text):
		"""
		<par>return a nicely formatted and colored string for
		the error text <arg>text</arg>.</par>
		"""
		return self.TextError(text)

	def strtarget(self, target):
		"""
		<par>return a nicely formatted and colored string for
		the target <arg>target</arg>.</par>
		"""
		try:
			name = target.__class__.name
		except AttributeError:
			name = target.__class__.__name__

		id = target.id
		if isinstance(id, url.URL) and id.islocal():
			if self.here is None:
				self.here = url.here()
			if self.home is None:
				self.home = url.home()
			s = str(id)
			test = str(id.relative(self.here))
			if len(test) < len(s):
				s = test
			test = "~/%s" % id.relative(self.home)
			if len(test) < len(s):
				s = test
		else:
			s = str(id)
		return self.TextTarget(name, "(", self.TextID(s), ")")

	def __setitem__(self, id, target):
		"""
		<par>set the target with the id <arg>name</arg>.</par>
		"""
		if id in self:
			self.reportWarning(RedefinedTargetWarning(id), 5)
		if isinstance(id, url.URL) and id.islocal():
			id = id.abs(scheme="file")
		super(Project, self).__setitem__(id, target)

	def __candidates(self, id):
		"""
		Return candidates for alternative forms of <arg>id</arg>.
		This is a generator, so when the first suitable candidate is use, the
		rest of the candidates won't have to be created at all.
		"""
		yield id
		if isinstance(id, basestring):
			id2 = url.URL(id)
			yield id2
			id2 = id2.abs(scheme="file")
			yield id2
			id2 = id2.real(scheme="file")
			yield id2
			if ":" in id:
				(prefix, rest) = id.split(":", 1)
				if prefix == "oracle":
					if "|" in rest:
						(connection, rest) = rest.split("|", 1)
						if ":" in rest:
							(type, name) = rest.split(":", 1)
							if "|" in rest:
								(name, key) = rest.split("|")
							else:
								key = None
							yield OracleID(connection, type, name, key)
		
	def __getitem__(self, id):
		"""
		<par>return the target with the id <arg>id</arg>.</par>
		<par>If an id can't be found, it will be wrapped in a
		<pyref module="ll.url" class="URL"><class>URL</class></pyref> instance and retried.</par>
		"""
		sup = super(Project, self)
		for id2 in self.__candidates(id):
			try:
				return sup.__getitem__(id2)
			except KeyError:
				pass
		raise UndefinedTargetError(id)

	def has_key(self, id):
		"""
		<par>Return whether the target with the id <arg>id</arg>
		exists in the project.</par>
		"""
		return id in self

	def __contains__(self, id):
		"""
		<par>Return whether the target with the id <arg>id</arg>
		exists in the project.</par>
		"""
		sup = super(Project, self)
		for id2 in self.__candidates(id):
			has = sup.has_key(id)
			if has:
				return True
		return False

	def has(self, id):
		warnings.warn("has() is deprecated, use has_key() or __contains__() (i.e. ... in ...) instead", DeprecationWarning)
		return self.has_key(id)

	def clear(self):
		"""
		<par>Removes all targets from <self/>. If you keep external references
		to targets they will be worthless, because the dependencies between
		target will have been removed.</par>
		"""
		for target in self.itervalues():
			target.clear()
		super(Project, self).clear()

	def create(self):
		"""
		<par>Create all dependencies for the project. Overwrite in subclasses.</par>

		<par>This method should only be called once, otherwise you'll get lots of
		<pyref class="RedefinedTargetWarning"><class>RedefinedTargetWarning</class>s</pyref>.
		But you can call <pyref method="clear"><method>clear</method></pyref> to
		remove all targets before calling <method>create</method>. You can also
		use the method <pyref method="recreate"><method>recreate</method></pyref> for that.</par>
		"""

	def recreate(self):
		"""
		<par>Calls <pyref method="clear"><method>clear</method></pyref> and
		<pyref method="create"><method>create</method></pyref> to recreate
		all project dependencies.</par>
		"""
		self.clear()
		self.create()

	def optionparser(self):
		"""
		Return an <module>optparse</module> parser from parsing the command line
		options. This can be overwritten in subclasses to add more options.
		"""
		p = optparse.OptionParser(usage="usage: %prog [options] [targets]", version="%%prog %s" % __version__)
		p.add_option("-v", "--verbosity", dest="verbosity", metavar="VERB", help="Verbosity of progress report: 'ioatds' lowercase activates, uppercase deactivates (i=inputs, o=outputs, l=level, a=actions, t=time, d=data, s=summary)", default="")
		p.add_option("-x", "--ignore", dest="ignoreerrors", help="Ignore errors", action="store_true", default=None)
		p.add_option("-X", "--noignore", dest="ignoreerrors", help="Don't ignore errors", action="store_false", default=None)
		p.add_option("-c", "--color", dest="color", help="Use colored output", action="store_true", default=None)
		p.add_option("-C", "--nocolor", dest="color", help="No colored output", action="store_false", default=None)
		return p

	def parseoptions(self, commandline=None):
		"""
		Use the parser returned by <pyref method="optionparser"><method>optionparser</method></pyref>
		to parse the option sequence <arg>commandline</arg>, modify <self/> accordingly and return
		the result of <module>optparse</module>s <method>parse_args</method> call.
		"""
		p = self.optionparser()
		(options, args) = p.parse_args(commandline)
		for c in options.verbosity:
			if c=="i":
				self.showinputs = True
			elif c=="I":
				self.showinputs = False
			elif c=="o":
				self.showoutputs = True
			elif c=="O":
				self.showoutputs = False
			elif c=="l":
				self.showlevel = True
			elif c=="L":
				self.showlevel = False
			elif c=="a":
				self.showactions = True
			elif c=="A":
				self.showactions = False
			elif c=="t":
				self.showtime = True
			elif c=="T":
				self.showtime = False
			elif c=="d":
				self.showdata = True
			elif c=="D":
				self.showdata = False
			elif c=="s":
				self.showsummary = True
			elif c=="S":
				self.showsummary = False
		if options.ignoreerrors is not None:
			self.ignoreerrors = options.ignoreerrors
		if options.color is not None:
			self.color = options.color
		return (options, args)

	def build(self, *targets):
		"""
		<par>build all targets in <arg>targets</arg>.</par>
		
		<par>Items in <arg>targets</arg> must be <pyref class="Target"><class>Target</class></pyref>
		instances or target ids.</par>

		<par>This method can be called multiple times and will recheck modification timestamp
		on each call.</par>
		"""
		t1 = dtnow()
		self.targetsrebuilt = 0
		self.targetsfailed = 0
		if self.__built:
			for target in self.itervalues():
				target.dirty()
		cache = url.importcache.copy() # remember the state of the import cache
		self.here = url.here()
		self.home = url.home()
		self.level = -1
		try:
			for target in targets:
				if not isinstance(target, Target):
					target = self[target]
				target.update()
			t2 = dtnow()
			if self.showsummary:
				self.report(
					self.TextProgress(
						"built ",
						self.TextTarget(self.__class__.__name__),
						"=",
						self.strseconds(t2-t1),
						" (",
						self.TextData(str(len(self))),
						" targets; ",
						self.TextData(str(self.targetsrebuilt)),
						" rebuilt; ",
						self.TextData(str(self.targetsfailed)),
						" failed)\n",
					)
				)
		finally:
			url.importcache = cache # restore the import cache
			self.__built = True # remember the call here, so targets will be set dirty on the next call
			self.here = None
			self.home = None
			self.level = None

	def reportphonytargets(self):
		"""
		Report a list of all <pyref class="PhonyTarget"><class>PythonTarget</class></pyref>
		and their documentation.
		"""
		self.report("Available phony targets are:\n")
		phonies = []
		maxlen = 0
		for id in self:
			if isinstance(id, basestring):
				maxlen = max(maxlen, len(id))
				phonies.append(self[id])
		phonies.sort(lambda p1, p2: cmp(p1.id, p2.id)) # FIXME: Use sort(key=...) in Python 2.4
		for phony in phonies:
			if phony.doc:
				id = ("%s %s" % (phony.id, "."*(maxlen+4)))[:maxlen+3] + " " # FIXME: Use str.rjust(fill=...) in Python 2.4
				doc = phony.doc
			else:
				id = phony.id
				doc = ""
			self.report("\t%s%s\n" % (id, doc))

	def buildwithargs(self, commandline=None):
		"""
		For calling make scripts from the command line. <arg>commandline</arg> defaults to <lit>sys.argv[1:]</lit>.
		"""
		if not commandline:
			import sys
			commandline = sys.argv[1:]
		(options, args) = self.parseoptions(commandline)

		if args:
			self.build(*args)
		else:
			self.reportphonytargets()

	def report(self, *texts):
		"""
		All screen output is done through this method. This makes it possible to redirect
		the output (e.g. to logfiles) in subclasses.
		"""
		sys.stderr.write("".join(map(str, texts)))
		sys.stderr.flush()

	def reportError(self, *texts):
		"""
		Report an error
		"""
		self.report(*texts)

	def reportWarning(self, warning, stacklevel):
		"""
		Issue a warning through the Python warnings framework
		"""
		warnings.warn(warning, stacklevel=stacklevel)
